/*
 * Copyright (C) 2022 European Spallation Source ERIC.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.openepics.names.docker;

import static org.junit.jupiter.api.Assertions.fail;

import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.UUID;

import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.openepics.names.rest.beans.Type;
import org.openepics.names.rest.beans.element.NameElement;
import org.openepics.names.rest.beans.element.NameElementCommand;
import org.openepics.names.rest.beans.element.StructureElement;
import org.openepics.names.rest.beans.element.StructureElementCommand;
import org.openepics.names.util.EnumUtil.NameChoice;
import org.testcontainers.containers.DockerComposeContainer;
import org.testcontainers.containers.wait.strategy.Wait;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Integration tests for Naming and PostgreSQL that make use of existing dockerization
 * with docker-compose.yml / Dockerfile.
 *
 * <p>
 * Focus of this class is names endpoint.
 * </p>
 *
 * @author Lars Johansson
 */
@Testcontainers
class NamesIT {

    // note
    //     NameElement extends NameElementCommand, StructureElement extends StructureElementCommand
    //     -->
    //     for such operations that send Command objects to backend, it is also possible to send
    //     NameElement, StructureElement

    @Container
    public static final DockerComposeContainer<?> ENVIRONMENT =
        new DockerComposeContainer<>(new File("docker-compose-it-db-schema-migration.yml"))
            .waitingFor(ITUtil.NAMING, Wait.forLogMessage(".*Started NamingApplication.*", 1));

    private static UUID systemGroupAcc  = null;
    private static UUID systemRFQ       = null;
    private static UUID subsystem010PRL = null;
    private static UUID subsystem010    = null;
    private static UUID subsystemN1U1   = null;

    private static UUID disciplineEMR   = null;
    private static UUID deviceGroupEMR  = null;
    private static UUID deviceTypeFS    = null;
    private static UUID deviceTypeRFA   = null;
    private static UUID deviceTypeTT    = null;

    @BeforeAll
    public static void initAll() {
        // init system group, system, subsystem, discipline, device group, device type

        StructureElementCommand structureElement  = null;
        StructureElement approvedStructureElement = null;

        structureElement = new StructureElementCommand(
                null, Type.SYSTEMGROUP, null,
                "Accelerator", "Acc",
                "The ESS Linear Accelerator", "approved by alfio");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        systemGroupAcc = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SYSTEM, systemGroupAcc,
                "Radio Frequency Quadrupole", "RFQ",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        systemRFQ = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SUBSYSTEM, systemRFQ,
                "01 Phase Reference Line", "010PRL",
                "empty", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        subsystem010PRL = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SUBSYSTEM, systemRFQ,
                "RFQ-010", "010",
                "empty", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        subsystem010 = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.SUBSYSTEM, systemRFQ,
                "Power switch board 01", "N1U1",
                "Electrical power cabinets", "Approved by Daniel Piso");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        subsystemN1U1 = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DISCIPLINE, null,
                "Electromagnetic Resonators", "EMR",
                "empty", "empty");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        disciplineEMR = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICEGROUP, disciplineEMR,
                "Control", null,
                "empty", "These names are needed now, so I am approving, but please add a description to these later.");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceGroupEMR = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICETYPE, deviceGroupEMR,
                "Flow Switch", "FS",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceTypeFS = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICETYPE, deviceGroupEMR,
                "RF Antenna", "RFA",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceTypeRFA = approvedStructureElement.getUuid();

        structureElement = new StructureElementCommand(
                null, Type.DEVICETYPE, deviceGroupEMR,
                "Temperature Transmitter", "TT",
                "empty", "Approve names added from misc device group");
        approvedStructureElement = ITUtilStructureElement.assertCreateApprove(structureElement);
        deviceTypeTT = approvedStructureElement.getUuid();
    }

    @Test
    void equivalenceName() {
        // purpose
        //     test name equivalence
        //
        // what
        //     read        equivalence name

        ITUtilNameElement.assertEquivalence("A2T-010PRL:RFS-PRLTap-071", "A2T-10PR1:RFS-PR1TAP-71");
    }

    @Test
    void existsName() {
        // purpose
        //     test if name exists
        //
        // what
        //     read        exists name

        ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
    }

    @Test
    void isLegacyName() {
        // purpose
        //     test if name is legacy name
        //
        // what
        //     read        is legacy name

        ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-071", Boolean.FALSE);
    }

    @Test
    void isValidToCreateName() {
        // purpose
        //     test if name is valid to create
        //
        // what
        //     read        is valid to create name

        ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-071", Boolean.TRUE);
    }

    @Test
    void checkCreate() {
        // purpose
        //     test conditions for create name
        //         not create itself
        //
        // what - combination of
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // validate create
        //         system structure
        //     or  system structure + device structure + index
        //     +
        //     description
        //     comment

        try {
            NameElementCommand nameElement = new NameElementCommand();

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-051", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-051", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-051", Boolean.TRUE);

            // validate create

            ITUtilNameElement.assertValidate("[{asdf]", NameChoice.CREATE, HttpURLConnection.HTTP_BAD_REQUEST);
            ITUtilNameElement.assertCreate("[{asdf]", HttpURLConnection.HTTP_BAD_REQUEST);

            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     subsystem
            // description
            // comment

            nameElement.setDescription("description");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElement.setComment("comment");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // convention name exists
            // already exists since create subsystem was approved
            nameElement.setParentsystemstructure(subsystem010PRL);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);

            // ----------
            // system structure + device structure + index
            //     subsystem, device type
            // description
            // comment

            nameElement.setIndex("051");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.TRUE);

            // ----------

            nameElement.setParentsystemstructure(null);
            nameElement.setIndex(null);
            nameElement.setParentdevicestructure(null);

            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     system
            // description
            // comment

            // convention name exists
            // already exists since create system was approved
            nameElement.setParentsystemstructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);

            // ----------
            // system structure + device structure + index
            //     system, device type
            // description
            // comment

            nameElement.setIndex("051");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.TRUE);

            // ----------

            nameElement.setParentsystemstructure(null);
            nameElement.setIndex(null);
            nameElement.setParentdevicestructure(null);

            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure
            //     system group
            // description
            // comment

            // convention name exists
            // already exists since create system group was approved
            nameElement.setParentsystemstructure(systemGroupAcc);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);

            // ----------
            // system structure + device structure + index
            //     system group, device type
            // description
            // comment

            nameElement.setIndex("051");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            nameElement.setParentdevicestructure(deviceTypeTT);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.TRUE);

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.TRUE);

            // ----------

            nameElement.setParentsystemstructure(null);
            nameElement.setIndex(null);
            nameElement.setParentdevicestructure(null);

            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
            ITUtilNameElement.assertCreate(nameElement, ITUtil.HTTP_UNPROCESSABLE_ENTITY);

            // ----------
            // system structure + device structure
            //     system, device type
            // description
            // comment

            // convention name exists
            // already exists since create system was approved
            nameElement.setParentsystemstructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);

            nameElement.setIndex("051");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.TRUE);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void create() {
        // purpose
        //     test create name
        //
        // what - combination of
        //     create      create names
        //     read        exists name
        //     read        is legacy name
        //     read        is valid to create
        //     read        validate create
        //
        // note
        //     with and without index

        try {
            ObjectMapper mapper = new ObjectMapper();

            NameElementCommand nameElement  = null;
            NameElementCommand nameElement2 = null;

            nameElement = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "052",
                    "description", "comment");

            nameElement2 = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "061",
                    "description", "comment");

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-052", Boolean.TRUE);

            ITUtilNameElement.assertValidate(
                    "[" + mapper.writeValueAsString(nameElement) + ","+mapper.writeValueAsString(nameElement2) +"]",
                    NameChoice.CREATE, HttpURLConnection.HTTP_OK, Boolean.TRUE, Boolean.TRUE);

            ITUtilNameElement.assertValidate(nameElement,  NameChoice.CREATE, Boolean.TRUE);
            ITUtilNameElement.assertValidate(nameElement2, NameChoice.CREATE, Boolean.TRUE);

            // create
            ITUtilNameElement.assertCreate(nameElement);

            ITUtilNameElement.assertExists("RFQ-010PRL:EMR-RFA-052", Boolean.TRUE);
            ITUtilNameElement.assertIsLegacy("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);
            ITUtilNameElement.assertIsValidToCreate("RFQ-010PRL:EMR-RFA-052", Boolean.FALSE);

            ITUtilNameElement.assertValidate(nameElement, NameChoice.CREATE, Boolean.FALSE);
        } catch (IOException e) {
            fail();
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkUpdate() {
        // purpose
        //     test conditions for update name
        //         not update
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //
        // note
        //     create in order to update
        //
        // validate update
        //     uuid
        //     +
        //         system structure
        //     or  system structure + device structure + index
        //     +
        //     description
        //     comment

        try {
            NameElementCommand nameElement = null;
            NameElement createdNameElement = null;

            nameElement = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "053",
                    "description", "comment");

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElement);

            // validate update

            nameElement.setUuid(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setUuid(createdNameElement.getUuid());
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setDescription(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setDescription("checkUpdate");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setComment(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setComment("checkUpdate");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            // ----------

            nameElement.setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setParentsystemstructure(systemGroupAcc);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setParentsystemstructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setParentsystemstructure(subsystemN1U1);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setParentsystemstructure(subsystem010PRL);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            // ----------

            nameElement.setParentdevicestructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            // convention name exists
            // already exists since create subsystem was approved
            nameElement.setIndex(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setIndex("053");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setParentdevicestructure(deviceGroupEMR);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setParentdevicestructure(deviceTypeFS);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setParentdevicestructure(deviceTypeTT);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setIndex(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            nameElement.setIndex("053");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            // system structure, device structure not used for validation
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void update() {
        // purpose
        //     test update name
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //     update      update names
        //
        // note
        //     create in order to update

        try {
            NameElementCommand nameElement = null;
            NameElement createdNameElement = null;

            nameElement = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "054",
                    "description", "comment");

            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.FALSE);

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElement);
            nameElement.setUuid(createdNameElement.getUuid());

            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            nameElement.setDescription("updated description");
            nameElement.setComment("updated comment");

            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);

            // update
            ITUtilNameElement.assertUpdate(nameElement);
            nameElement.setDescription("another description");
            nameElement.setComment("another comment");

            ITUtilNameElement.assertValidate(nameElement, NameChoice.UPDATE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void checkDelete() {
        // purpose
        //     test conditions for delete name
        //         not delete
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate delete
        //
        // note
        //     create in order to delete
        //
        // validate delete
        //     uuid
        //     comment

        try {
            NameElementCommand nameElement = null;
            NameElement createdNameElement = null;

            nameElement = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "055",
                    "description", "comment");

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElement);
            nameElement.setUuid(createdNameElement.getUuid());

            // validate delete

            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setDescription(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setDescription("checkDelete");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setComment(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.FALSE);

            nameElement.setComment("checkDelete");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            // system structure not used for validation

            nameElement.setParentsystemstructure(systemGroupAcc);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setParentsystemstructure(systemRFQ);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setParentsystemstructure(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setParentsystemstructure(subsystem010PRL);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            // device type not used for validation

            nameElement.setParentdevicestructure(deviceTypeRFA);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            // index not used for validation

            nameElement.setIndex(null);
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setIndex("055");
            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void delete() {
        // purpose
        //     test delete name
        //
        // what - combination of
        //     create      create names
        //     read        validate create
        //     read        validate update
        //     delete      delete names
        //
        // note
        //     create in order to delete

        try {
            NameElementCommand nameElement = null;
            NameElement createdNameElement = null;
            NameElement deletedNameElement = null;

            nameElement = new NameElementCommand(
                    null, subsystem010PRL, deviceTypeRFA, "056",
                    "description", "comment");

            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.FALSE);

            // create
            createdNameElement = ITUtilNameElement.assertCreate(nameElement);
            nameElement = createdNameElement;

            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            nameElement.setDescription("deleted description");
            nameElement.setComment("deleted comment");

            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.TRUE);

            // delete
            deletedNameElement = ITUtilNameElement.assertDelete(nameElement);
            nameElement = deletedNameElement;

            ITUtilNameElement.assertValidate(nameElement, NameChoice.DELETE, Boolean.FALSE);
        } catch (Exception e) {
            fail();
        }
    }

    @Test
    void readSearchHistoryDeleted() {
        // purpose
        //     test read names in various ways for create
        //         search
        //       ( latest )
        //         history
        //         deleted
        //
        // note
        //     create (and more) to read (with content)

        NameElementCommand nameElement  = null;
        NameElement responseNameElement = null;
        UUID uuid, uuid2 = null;

        NameElement nameElement1, nameElement7, nameElement8 = null;

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "001",
                "description", "comment");

        // create
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);
        nameElement1 = responseNameElement;
        uuid = responseNameElement.getUuid();

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "002",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);
        uuid2 = responseNameElement.getUuid();

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "003",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "004",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "005",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "006",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "007",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);
        nameElement7 = responseNameElement;

        nameElement = new NameElementCommand(
                null, subsystem010, deviceTypeFS, "008",
                "description", "comment");
        responseNameElement = ITUtilNameElement.assertCreate(nameElement);
        nameElement8 = responseNameElement;

        // update element 1 twice
        nameElement = nameElement1;
        nameElement.setDescription("updated description");
        nameElement.setComment("updated comment");

        responseNameElement = ITUtilNameElement.assertUpdate(nameElement);
        nameElement = responseNameElement;

        nameElement.setDescription("updated description again");
        nameElement.setComment("updated comment again");

        responseNameElement = ITUtilNameElement.assertUpdate(nameElement);
        nameElement = responseNameElement;

        // delete element 7, 8
        responseNameElement = ITUtilNameElement.assertDelete(nameElement7);
        responseNameElement = ITUtilNameElement.assertDelete(nameElement8);

        try {
            // read & search
            //     deleted, uuid, name, nameequivalence, systemstructure, devicestructure, index, description
            //     combination
            ITUtilNameElement.assertRead("", 8, -1);

            ITUtilNameElement.assertRead("?deleted=false", 6, -1);
            ITUtilNameElement.assertRead("?deleted=true",  2, -1);

            ITUtilNameElement.assertRead("?uuid=" + deviceTypeFS.toString(), 0);
            ITUtilNameElement.assertRead("?uuid=" + uuid.toString(),         1);

            ITUtilNameElement.assertRead("?name=RFQ-010:EMR-FS-005",  1);
            ITUtilNameElement.assertRead("?name=RFQ-010%",            8, -1);
            ITUtilNameElement.assertRead("?name=RFQ-10%",             0);

            ITUtilNameElement.assertRead("?nameequivalence=RFQ-10%",  8, -1);

            ITUtilNameElement.assertRead("?systemstructure=RFQ-010",  6, -1);
            ITUtilNameElement.assertRead("?systemstructure=RFQ-0",    0);
            ITUtilNameElement.assertRead("?systemstructure=RFQ-0__",  6, -1);
            ITUtilNameElement.assertRead("?systemstructure=RFQ-N1U1", 1);

            ITUtilNameElement.assertRead("?devicestructure=EMR-FS",   6, -1);
            ITUtilNameElement.assertRead("?devicestructure=EMR-F",    0);
            ITUtilNameElement.assertRead("?devicestructure=EMR-F_",   6, -1);
            ITUtilNameElement.assertRead("?devicestructure=EMR-TT",   0);

            ITUtilNameElement.assertRead("?index=003", 1);

            ITUtilNameElement.assertRead("?description=description",               7, -1);
            ITUtilNameElement.assertRead("?description=%description%",             8, -1);
            ITUtilNameElement.assertRead("?description=updated description",       0, -1);
            ITUtilNameElement.assertRead("?description=updated description%",      1, -1);
            ITUtilNameElement.assertRead("?description=updated description again", 1);

            ITUtilNameElement.assertRead("?nameequivalence=RFQ-10%&devicestructure=EMR-FS", 6, -1);

            // name or uuid
            //     /{name}
            ITUtilNameElement.assertRead("/" + systemGroupAcc.toString(), 0);
            ITUtilNameElement.assertRead("/" + systemRFQ.toString(),      0);
            ITUtilNameElement.assertRead("/" + subsystem010.toString(),   0);
            ITUtilNameElement.assertRead("/" + disciplineEMR.toString(),  0);
            ITUtilNameElement.assertRead("/" + deviceGroupEMR.toString(), 0);
            ITUtilNameElement.assertRead("/" + deviceTypeFS.toString(),   0);
            ITUtilNameElement.assertRead("/" + uuid.toString(),           1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-001",           1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0",             0);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0__",           6, -1);
            ITUtilNameElement.assertRead("/RFQ-010:EMR-FS-0%",            6, -1);

            // system structure mnemonic path
            //     /systemstructure/{mnemonicpath}
            ITUtilNameElement.assertRead("/systemstructure/RFQ-010",      6, -1);
            ITUtilNameElement.assertRead("/systemstructure/RFQ-0",        0);
            ITUtilNameElement.assertRead("/systemstructure/RFQ-0__",      6, -1);
            ITUtilNameElement.assertRead("/systemstructure/RFQ-N1U1",     1);

            // device structure mnemonic path
            //     /devicestructure/{mnemonicpath}
            ITUtilNameElement.assertRead("/devicestructure/EMR-FS",       6, -1);
            ITUtilNameElement.assertRead("/devicestructure/EMR-F",        0);
            ITUtilNameElement.assertRead("/devicestructure/EMR-F_",       6, -1);
            ITUtilNameElement.assertRead("/devicestructure/EMR-TT",       0);

            // history
            //     /history/{uuid}
            ITUtilNameElement.assertRead("/history/" + systemGroupAcc.toString(), 0);
            ITUtilNameElement.assertRead("/history/" + systemRFQ.toString(),      0);
            ITUtilNameElement.assertRead("/history/" + subsystem010.toString(),   0);
            ITUtilNameElement.assertRead("/history/" + disciplineEMR.toString(),  0);
            ITUtilNameElement.assertRead("/history/" + deviceGroupEMR.toString(), 0);
            ITUtilNameElement.assertRead("/history/" + deviceTypeFS.toString(),   0);
            ITUtilNameElement.assertRead("/history/" + uuid.toString(),           3);
            ITUtilNameElement.assertRead("/history/" + uuid2.toString(),          1);
        } catch (Exception e) {
            fail();
        }
    }

}
